// ALDL_Communication 8192 baud
//
// Mike Waldron 2013
// http://www.xm381.com/xm381/ALDL.html has ALDL info and documentation
//
// This Arduino sketch will transmit a "Mode 1" ALDL requests (8192 baud only) to an GM ECU
// and validate the ECU response, with valid ALDL data stored in the array ALDLStream[].
// The sketch is configured for a $A1 (BFUZ) prom on a 1227727 ECU.
// It can be made to work with any 8192 ECU by entering the ECUs configuration information.
// If your ECU generates "chatter", you will need to add the commands to turn chatter off. Use beginMode1() as a template. 

// ***** $A1 ECU Configuration Info *********************************************************************************************
const byte Mode1Request[] = { 0xF4, 0x57, 0x01, 0x00 }; // Mode 1 Request byte string 
const byte Mode1Header[] = { 0xF4, 0x95, 0x01 };        // This is the header returned with the ALDL data
const byte Mode1DataBytes = 63;                         // This is the number of data bytes returned (not including the header)

// Status Bytes: ValueName = byte position in ALDLStream ( must offset location in ALDLStreamBuffer, offset = Mode1RequestSize + Mode1HeaderSize )  

const byte CTS = 7;   // (COOLANT TEMPERATURE, deg F = 1.35N - 40)
const byte TPS = 10;  // *0.391 (RESCALED TPS, % THROTTLE = N/2.56)
const byte RPM = 11;  // *25 (ENGINE RPM)
const byte TBRP = 14; // (TIME BETWEEN REFERENCE PULSES (MSB), 2 byte, mSEC = (MSB*256 + LSB)/65.536)
const byte MPH = 17;  // (FILTERED MILES PER HOUR)
const byte O2 = 19;   // (O2 mV = N * 4.42)
const byte BLM = 22;  // (BLOCK LEARN MULTIPLIER, 128 perfect, lower number leans too rich mixter, higher number enrichens too lean mixture)
const byte BLC = 23;  // (CURRENT BLOCK LEARN MULTIPLIER CELL)
const byte IAC = 25;  // (IAC PRESENT MOTOR POSITION)
const byte DIS = 27;  // (DESIRED IDLE SPEED, RPM = N*12.5)
const byte MAP = 29;  // *0.313 (MANIFOLD ABSOLUTE PRESSURE, kpa = (N + 28.06)/2.71)
const byte WB = 31;   // *0.2 + 9.7 (wideband O2 connected to "fuel pump volt" input of ECU, volts = /10, A/F = volts/2 + 9, + .7 correction for my installation)
const byte PMP = 31;  // (FUEL PUMP RELAY, VOLTS = N/10)
const byte BAT = 34;  // (BATTERY VOLTAGE, VOLTS = N/10)
const byte AFG = 37;  // (ENGINE AIRFLOW, GRAMS/SECOND = N)
const byte SPK = 38;  // (SPARK ADVANCE REL TO TDC (MSB) 2 byte, 2's complement data *0.352 [65536 - ((MSB * 256)+(LSB * 0.3515625))]
const byte SPK1 = 40; // *0.352 - (SPARK ADVANCE, DEGREES = N*90/256 Values seem suspect, calc wrong?) x * .351565
const byte AF = 41;   // *0.1 (DESIRED AIR/FUEL RATIO, A/F = N/10)
const byte BPW = 42;  // (FINAL BASE PULSE WIDTH (MSB), 2 byte, mSEC = (MSB*256 + LSB)/65.536)
const byte RTM = 48;  // (ENGINE RUNNING TIME IN SECONDS (MSB), 2 byte, SECONDS = MSB*256 + LSB)
const byte CNV = 50;  // (CATALYTIC CONVERTER TEMPERATURE (estimated by cpu, no sensor) DEGREES C = N*3 + 300)




// Status Bits: ValueName[] = {ALDL data byte, bit in data byte, true when bit is}
const byte Rich[] = {63,6};          // (1 = RICH 0 = LEAN) 
const byte OpenLoop[] = {63,7};      // (1 = IN CLOSED LOOP, 0 = OPEN LOOP)
const byte Idle[] = {55,7};          // (1 = IDLE CONDITIONS MET)
const byte ThrottleClosed[] = {62,7};// (1 = CLOSED TPS FLAG)
const byte HighwayMode[] = {56,3};   // (1 = HIGHWAY SPARK ADVANCE (CONVERTER OVERTEMP MODE))
const byte DFCO[] = {62,3};          // (DECEL FUEL CUTOFF FLAG)
const byte MALF[] = {58,1};          // (1 = MALFS HAVE OCCURRED)
const byte TCC[] = {52,3};           // (1 = TCC LOCKED / SHIFT LIGHT ON)

// More Status Bits
const byte FN1[] = {18,4};           //(1 = FAN #1 TO BE TURNED ON BY OUTPUT LOGIC)
const byte FN2[] = {18,5};           //(1 = FAN #2 TO BE TURNED ON BY OUTPUT LOGIC)
const byte ACC[] = {54,5};           //(0 = A/C CLUTCH ENABLED)
const byte EGR[] = {56,4};           //(1 = EGR ON)
const byte LMA[] = {58,2};           //(1 = LAUNCH MODE)
const byte ACR[] = {59,7};           //(AIR CONDITIONER REQUEST, 1=ON   0=OFF)
const byte IMF[] = {62,5};           //(IDLE MODE FUEL FLAG, 1 = CALCULATE BPW WITH AIRFLOW)
const byte LRN[] = {63,1};           //(LEARN CONTROL ENABLE FLAG, 1=ENABLE)


// Malfunctions Status Bytes with "value" for each bit in an array
const byte MALFFLG1byte = 3;
const char* MALFFLG1 [] = { 
	"VSS LOW",
	"MAT LOW",
	"TPS LOW",
	"TPS HI",
	"CTS LOW",
	"CTS HIGH",
	"O2 SENSOR",
	"REF PULSE",
};
const byte MALFFLG2byte = 4;
const char* MALFFLG2[] = {
	"EST ERR",
	"CYLSEL ERR",
	"IAC ERR",
	"MAP LOW",
	"MAP HIGH",
	"EVRV/EGR",
	"INJ MALF",
	"MAT HI",
};
const byte MALFFLG3byte = 5;
const char* MALFFLG3[] = {
	"ADU ERR",
	"PUMP V LOW",
	"BAT HIGH",
	"",
	"PROM ERR",
	"O2 RICH",
	"O2 LEAN",
	"ESC ERR",
};
const byte MALFFLG4byte = 6;
const char* MALFFLG4[] = {
	"",
	"A/C SENSOR",
	"",
	"",
	"TRANS SW",
	"VATS ERR",
	"DEGRADE O2",
	"",
};


// ******************************************************************************************************************************

// ***** Hardware Config *****
//ALDL
#define ALDLSerial Serial

// ***** Variables for ALDL data *****

// Calculate how many bytes long is the Mode 1 request string.
const byte Mode1RequestSize = sizeof(Mode1Request)/sizeof(Mode1Request[0]);

// Calculate how many bytes long is the Mode 1 header string. 
const byte Mode1HeaderSize = sizeof(Mode1Header)/sizeof(Mode1Header[0]); 

// # of Request bytes + checksum byte + header bytes + data bytes + checksum byte
const byte ALDLStreamSize = (Mode1RequestSize + 1 + Mode1HeaderSize + Mode1DataBytes + 1);  

//starting location of ALDL in buffer (after request and header data)
byte offset = Mode1RequestSize + Mode1HeaderSize; 
byte ALDLStream[ALDLStreamSize]; // Array for verified ALDL data
byte ALDLStreamBuffer[ALDLStreamSize]; // Array to store ALDL for verification 
byte checkSum;
byte Mode1RequestcheckSumByte; // Checksum of Mode 1 bytes
byte ALDLValidity;
unsigned long timeout;

// ***** Program Setup *********************************************************************************
void setup() {
  ALDLSerial.begin(8192); 

// ***** Mode 1 checksum calculation *****
  checkSum = 0;
  for (byte Requestbytes=0; Requestbytes < Mode1RequestSize - 1; Requestbytes++) {
    checkSum += Mode1Request[Requestbytes];
  }
  Mode1RequestcheckSumByte = 0 - checkSum;
  
}

// ***** Main Loop *********************************************************************************
void loop() {
  
    // request mode 1
      beginMode1();
    
    // get ALDL data
      getALDLStream(); 

    // Verify ALDLStream    
      VerifyALDLStream();
  
}

// ***** Mode 1 Request *********************************************************************************
void beginMode1() {
   
  while(ALDLSerial.available() != 0){ // empty serial buffer
    byte bitbucket = ALDLSerial.read();
  }
  
  for(byte bytesRead = 0; bytesRead <= Mode1RequestSize - 1; bytesRead++) { // send request  
    ALDLSerial.write (Mode1Request[bytesRead]);
  }
  ALDLSerial.write (Mode1RequestcheckSumByte); // send checksum
}

// ***** getALDLStream *****
void getALDLStream() {
  
  for (byte Clearbytes=0; Clearbytes < ALDLStreamSize; Clearbytes++) { // empty ALDL Stream array
    ALDLStream[Clearbytes] = 0x00;
  }
        
  for (byte Clearbytes=0; Clearbytes < ALDLStreamSize; Clearbytes++) { // empty ALDL Stream Buffer array
    ALDLStreamBuffer[Clearbytes] = 0x00;
  }
  
  byte bytesRead = 0;      
  timeout = millis();
  
  while((millis() - timeout) < 100){
    if (ALDLSerial.available()) { // retrieve ALDL response data from ECU in buffer, if any
      byte dataByte = ALDLSerial.read();
      ALDLStreamBuffer[bytesRead] = dataByte;
//      BluetoothSerial.write (dataByte);
      bytesRead++;
    }
    if(bytesRead >= ALDLStreamSize){
      break;
    } 
  }
}

// verify ALDL data stream: Mode1Request + checksum + header + ALDL data + checksum byte should equal 0               
// ***** Verify ALDLStream *********************************************************************************
void VerifyALDLStream(){
  // check header string in ALDL data stream: does it match Mode1Header string        
  byte Mode1HeadercheckSum=0;
       checkSum=0;

  for (byte Requestbytes=0; Requestbytes < Mode1HeaderSize; Requestbytes++) {
    Mode1HeadercheckSum += Mode1Header[Requestbytes];
    checkSum += ALDLStreamBuffer[Requestbytes + Mode1RequestSize + 1];
  }

  if (Mode1HeadercheckSum != checkSum) {
    ALDLValidity = 2;
    return;
  }

  // check ALDL data stream: header + ALDL data + checksum byte should equal 0               
  checkSum=0;
  
  for (byte Requestbytes=0; Requestbytes < ALDLStreamSize - (Mode1RequestSize + 1); Requestbytes++) {
    checkSum += ALDLStreamBuffer[Requestbytes + (Mode1RequestSize + 1)];
  }

  if (checkSum != 0) {
    ALDLValidity = 1;
    return;
  }

  if (checkSum == 0) {
  ALDLValidity = 0;
  
    for(byte bytesRead = 0; bytesRead < ALDLStreamSize ; bytesRead++){
      ALDLStream[bytesRead] = ALDLStreamBuffer[bytesRead]; // transfer good data to ALDLStream     
    }
  } 
}





